"""
Created on: 26-1-2023 12:33

@author: IvS
"""

import pytest
import numpy as np

from pydsol.model.source import Source
from pydsol.model.server import Server


def test_source_class(mocker):
    mock_simulator = mocker.Mock()
    source = Source(mock_simulator)

    assert source.simulator == mock_simulator
    assert source.id == 1
    assert isinstance(source.id, int)
    assert source.name == "Source 1"
    assert isinstance(source.name, str)

    mock_entity_type = mocker.Mock()
    source_2 = Source(mock_simulator, interarrival_time=1, num_entities=3, entity_type=mock_entity_type, name="test",
                      distribution=np.random.exponential)
    assert source_2.id == 2
    assert source_2.interarrival_time == 1
    assert source_2.num_entities == 3
    assert source_2.entity_type == mock_entity_type
    assert source_2.name == "test"
    assert source_2.distribution == np.random.exponential


def test_create_entities(mocker):
    mock_simulator = mocker.Mock()
    mock_simulator.simulator_time = 1

    # Test with default interarrival time and default entity
    source = Source(mock_simulator)
    source.create_entities()

    # Test with multiple entities
    source = Source(mock_simulator, num_entities=3)
    source.create_entities()

    # Test with other interarrival time
    source = Source(mock_simulator, interarrival_time=1)
    source.create_entities()

    #None-iterable distribution
    source = Source(mock_simulator, interarrival_time=1, distribution=np.random.normal, num_entities=5)
    source.create_entities()

    # Test with other interarrival time and distribution
    source = Source(mock_simulator, interarrival_time=(0, 1, 2), distribution=np.random.triangular)
    source.create_entities()

    # Test with all variables
    source = Source(mock_simulator, interarrival_time=(0, 1, 2), distribution=np.random.triangular, num_entities=5)
    source.create_entities()

    # Not matching distribution and transfer_in_time
    with pytest.raises(TypeError):
        source = Source(mock_simulator, interarrival_time=1, distribution=np.random.triangular)
        source.create_entities()


def test_enter_output_node(mocker):
    mock_simulator = mocker.Mock()
    mock_simulator.simulator_time = 1

    mock_entity = mocker.Mock()
    mock_entity.name = "test"

    vehicle = mocker.Mock()
    vehicle.name = "vehicle"
    vehicle.entities_on_vehicle = []

    # With vehicle
    source = Source(mock_simulator, vehicle_type=vehicle)
    source.enter_output_node(mock_entity)
    assert isinstance(source.kwargs["vehicle_type"], object)

    # With wrong vehicle unit
    with pytest.raises(TypeError):
        vehicle_type = 1
        source = Source(mock_simulator, vehicle_type=vehicle_type)
        source.enter_output_node(mock_entity)

    # With vehicle speed
    vehicle_speed = 10
    source = Source(mock_simulator, vehicle_type=vehicle, vehicle_speed=vehicle_speed)
    source.enter_output_node(mock_entity)

    assert isinstance(vehicle_speed, (int, float))


def test_exit_output_node(mocker):
    mock_simulator = mocker.Mock()
    mock_entity = mocker.Mock()

    server_next = Server(mock_simulator)

    source = Source(mock_simulator)

    # No next
    with pytest.raises(AttributeError):
        source.exit_output_node(mock_entity)

    # Test with only one next node
    source.next = server_next
    source.exit_output_node(mock_entity)

    # Test with two next nodes
    server_1 = Server(mock_simulator)
    source.next = [server_next, server_1]
    source.exit_output_node(mock_entity)

    # Test with two nodes and selection weight
    server_next.selection_weight = 0.5
    server_1.selection_weight = 0.9

    source.exit_output_node(mock_entity)


if __name__ == '__main__':
    pytest.main()



"""
Created on: 26-1-2023 14:07

@author: IvS
"""

import pytest

from pydsol.model.link import Link, TimePath


def test_link_class(mocker):
    mock_simulator = mocker.Mock()
    mock_origin = mocker.Mock()
    mock_destination = mocker.Mock()

    mock_length = 5
    assert isinstance(mock_length, (int, float))

    link = Link(mock_simulator, mock_origin, mock_destination, mock_length)

    assert link.simulator == mock_simulator
    assert link.origin == mock_origin
    assert link.destination == mock_destination
    assert link.length == mock_length
    assert link.selection_weight == 1
    assert link.id == 1
    assert link.name == "Link 1"

    link_2 = Link(mock_simulator, mock_origin, mock_destination, mock_length, selection_weight=2, name="test", test=0)

    assert link_2.id == 2
    assert link_2.selection_weight == 2
    assert link_2.name == "test"
    assert link_2.kwargs == dict(name="test", test=0)


def test_link_enter_input_node(mocker):
    mock_simulator = mocker.Mock()
    mock_simulator.simulator_time = 1

    mock_origin = mocker.Mock()
    mock_origin.name = "origin"

    mock_destination = mocker.Mock()
    mock_destination.name = "destination"

    mock_length = 5

    link = Link(mock_simulator, mock_origin, mock_destination, mock_length)

    mock_entity = mocker.Mock()
    mock_entity.speed = 10
    mock_entity.name = "test"

    link.enter_input_node(mock_entity)


def test_link_exit_link(mocker):
    mock_simulator = mocker.Mock()
    mock_simulator.simulator_time = 1

    mock_origin = mocker.Mock()
    mock_origin.name = "origin"

    mock_destination = mocker.Mock()
    mock_destination.name = "destination"

    mock_length = 5

    link = Link(mock_simulator, mock_origin, mock_destination, mock_length)
    link_2 = Link(mock_simulator, mock_origin, mock_destination, 10)

    link.next = link_2

    mock_entity = mocker.Mock()
    mock_entity.speed = 10
    mock_entity.name = "test"

    link.exit_link(mock_entity)


def test_timepath_class(mocker):
    mock_simulator = mocker.Mock()
    mock_origin = mocker.Mock()
    mock_destination = mocker.Mock()

    mock_time = 5
    assert isinstance(mock_time, (int, float))

    timepath = TimePath(mock_simulator, mock_origin, mock_destination, mock_time)

    assert timepath.simulator == mock_simulator
    assert timepath.origin == mock_origin
    assert timepath.destination == mock_destination
    assert timepath.time == mock_time
    assert timepath.selection_weight == 1
    assert timepath.id == 1
    assert timepath.name == "TimePath 1"

    timepath_2 = TimePath(mock_simulator, mock_origin, mock_destination, mock_time, selection_weight=2, name="test",
                          test=0)

    assert timepath_2.id == 2
    assert timepath_2.selection_weight == 2
    assert timepath_2.name == "test"
    assert timepath_2.kwargs == dict(name="test", test=0)


def test_timepath_enter_input_node(mocker):
    mock_simulator = mocker.Mock()
    mock_simulator.simulator_time = 1

    mock_origin = mocker.Mock()
    mock_origin.name = "origin"

    mock_destination = mocker.Mock()
    mock_destination.name = "destination"

    mock_time = 5

    timepath = TimePath(mock_simulator, mock_origin, mock_destination, mock_time)

    mock_entity = mocker.Mock()
    mock_entity.name = "test"

    timepath.enter_input_node(mock_entity)


def test_timepath_exit_path(mocker):
    mock_simulator = mocker.Mock()
    mock_simulator.simulator_time = 1

    mock_origin = mocker.Mock()
    mock_origin.name = "origin"

    mock_destination = mocker.Mock()
    mock_destination.name = "destination"

    mock_time = 5

    timepath = TimePath(mock_simulator, mock_origin, mock_destination, mock_time)
    timepath_2 = TimePath(mock_simulator, mock_origin, mock_destination, mock_time)

    timepath.next = timepath_2

    mock_entity = mocker.Mock()
    mock_entity.name = "test"

    timepath.exit_path(mock_entity)


if __name__ == '__main__':
    pytest.main()



"""
Created on: 30-1-2023 13:01

@author: IvS
"""
import pytest
import numpy as np

from pydsol.model.server import Server
from pydsol.model.entities import Vehicle


def test_server_class(mocker):
    mock_simulator = mocker.Mock()

    server = Server(mock_simulator)

    assert server.simulator == mock_simulator
    assert isinstance(server.capacity, int)
    assert server.id == 1
    assert isinstance(server.id, int)
    assert server.name == "Server 1"
    assert isinstance(server.name, str)
    assert len(server.resources) == server.capacity
    assert any(isinstance(resource, object) for resource in server.resources)
    assert isinstance(server.input_queue.contents, list)

    server_2 = Server(mock_simulator, capacity=2, distribution=np.random.normal, processing_time=1, transfer_in_time=1,
                      name="test")

    assert server_2.id == 2
    assert server_2.capacity == 2
    assert server_2.distribution == np.random.normal
    assert server_2.processing_time == 1
    assert server_2.transfer_in_time == 1
    assert server_2.name == "test"
    assert isinstance(server_2.name, str)
    assert len(server_2.resources) == server_2.capacity
    assert any(isinstance(resource, object) for resource in server_2.resources)


def test_enter_input_node(mocker):
    mock_simulator = mocker.Mock()
    mock_simulator.simulator_time = 1

    mock_entity = mocker.Mock()
    mock_entity.name = "entity"

    mock_vehicle = Vehicle(mock_simulator)

    server = Server(mock_simulator)

    # Entity is not vehicle
    server.enter_input_node(mock_entity)

    # Entity is vehicle but wrong entities on vehicle
    with pytest.raises(AttributeError):
        mock_vehicle.entities_on_vehicle = ["1", "2", "3"]
        server.enter_input_node(mock_vehicle)

    # Entity is vehicle with right entites on vehicle
    mock_entity_1 = mocker.Mock(name="entity1")
    mock_entity_2 = mocker.Mock(name="entity2")
    mock_vehicle.entities_on_vehicle = [mock_entity, mock_entity_1, mock_entity_2]
    server.enter_input_node(mock_vehicle)

    # #Entity is vehicle with other data format
    mock_vehicle.entities_on_vehicle = (mock_entity, mock_entity_1, mock_entity_2)
    server.enter_input_node(mock_vehicle)


def test_seize_resource(mocker):
    mock_simulator = mocker.Mock(simulator_time=1)
    mock_entity = mocker.Mock(name="entity")

    server = Server(mock_simulator)

    mock_resource_1 = mocker.Mock(resource_seized=True, name="resource 1")
    mock_resource_2 = mocker.Mock(resource_seized=True, name="resource 2")
    mock_resource_3 = mocker.Mock(resource_seized=True, name="resource 3")

    # Test when all are taken that entity goes to queue
    server.resources = [mock_resource_1, mock_resource_2, mock_resource_3]
    server.seize_resource(mock_entity)
    assert len(server.input_queue.contents) == 1

    # Test when only one resource and second entity
    mock_entity_2 = mocker.Mock(name="entity2")
    server.resources = [mock_resource_1]
    server.seize_resource(mock_entity_2)
    assert len(server.input_queue.contents) == 2

    # Set resource not seized and check if entity is the first in line
    mock_resource_1.resource_seized = False
    mock_entity_3 = mocker.Mock(name="entity3")
    server.seize_resource(mock_entity_3)
    assert mock_resource_1.processing_entity == mock_entity
    assert mock_resource_1.resource_seized
    assert len(server.input_queue.contents) == 2

    # Test when input queue is zero and multiple resources
    mock_resource_1.resource_seized = mock_resource_2.resource_seized = mock_resource_3.resource_seized = False
    server.resources = [mock_resource_1, mock_resource_2, mock_resource_3]
    server.input_queue.contents = []
    server.seize_resource(mock_entity)
    assert mock_resource_1.resource_seized
    assert mock_resource_1.processing_entity == mock_entity

    # Test when first resource is taken
    mock_resource_1.resource_seized = True
    mock_resource_2.resource_seized = mock_resource_3.resource_seized = False
    server.resources = [mock_resource_1, mock_resource_2, mock_resource_3]
    server.input_queue.contents = []
    server.seize_resource(mock_entity)
    assert mock_resource_2.resource_seized
    assert mock_resource_2.processing_entity == mock_entity


def test_enter_output_node(mocker):
    mock_simulator = mocker.Mock(simulator_time=1)
    mock_entity = mocker.Mock(name="entity")
    mock_vehicle = mocker.Mock(name="vehicle", entities_on_vehicle=[])

    # With vehicle
    server = Server(mock_simulator, vehicle_type=mock_vehicle)
    server.enter_output_node(mock_entity)
    assert isinstance(server.kwargs["vehicle_type"], object)

    # With wrong vehicle unit
    with pytest.raises(TypeError):
        vehicle_type = 1
        server = Server(mock_simulator, vehicle_type=vehicle_type)
        server.enter_output_node(mock_entity)

    # With wrong vehicle unit
    with pytest.raises(TypeError):
        vehicle_type = "vehicle"
        server = Server(mock_simulator, vehicle_type=vehicle_type)
        server.enter_output_node(mock_entity)

    # With vehicle speed
    server = Server(mock_simulator, vehicle_type=mock_vehicle, vehicle_speed=10)
    server.enter_output_node(mock_entity)
    assert isinstance(server.kwargs["vehicle_speed"], (int, float))


def test_exit_output_node(mocker):
    mock_simulator = mocker.Mock()
    mock_entity = mocker.Mock()

    next_1 = Server(mock_simulator)
    next_2 = Server(mock_simulator)

    server = Server(mock_simulator)

    # No next
    with pytest.raises(AttributeError):
        server.exit_output_node(mock_entity)

    # Test with only one next node
    server.next = next_1
    server.exit_output_node(mock_entity)

    # Test with two next nodes
    server.next = [next_1, next_2]
    server.exit_output_node(mock_entity)

    # Test with two nodes and selection weight
    next_1.selection_weight = 0.5
    next_2.selection_weight = 0.9
    server.exit_output_node(mock_entity)


if __name__ == '__main__':
    pytest.main()



""" """
"""
Created on: 22-7-2021 10:19

@author: IvS
"""
import numpy as np
import itertools

from pydsol.model.entities import Entity
from pydsol.model.basic_logger import get_module_logger

logger = get_module_logger(__name__)

__all__ = ["Source"]


class Source(object):
    """This class defines a basic source for a discrete event simulation model. A Source is the start-station
    of each entity, meaning the entities enter the system via the source. Entities are created at a source,
    given a specific interarrival time."""
    id_iter = itertools.count(1)

    def __init__(self, simulator, interarrival_time="default", num_entities=1, entity_type=Entity, **kwargs):
        """

        Parameters
        ----------
        simulator: Simulator object
            simulator of the model.
        interarrival_time:
            time between the creation of entities. Default is np.random.exponential(0.25).
        num_entities: int
            number of entities to create at once. Default is 1.
        kwargs:
            kwargs are the keyword arguments that are used to expand the source class.
            *name: str
                user-specified source for the link
        """
        self.simulator = simulator
        self.num_entities = num_entities
        self.entity_type = entity_type

        self.next = None  # this should be the next process

        self.id = next(self.id_iter)
        self.name = "{0} {1}".format(self.__class__.__name__, str(self.id))
        if "name" in kwargs:
            self.name = kwargs["name"]

        self.interarrival_time = interarrival_time
        if "distribution" in kwargs:
            self.distribution = kwargs["distribution"]

        self.kwargs = kwargs

        self.simulator.schedule_event_now(self, "create_entities")

    def create_entities(self, **kwargs):
        """
        Create entities via SimEvent, given the interarrival time and the number of entities.

        Parameters
        ----------
        entity_type: class
            class where to make instances of, for example class Entity
        kwargs:
            kwargs are the keyword arguments that are used to invoke the method or expand the function.

        """

        # To make it work with the set seed of the simulator
        if self.interarrival_time == "default":
            self.interarrival_time = np.random.exponential(0.25)

        # Create new entity
        for _ in range(self.num_entities):
            entity = self.entity_type(self.simulator, self.simulator.simulator_time, **kwargs)
            logger.info("Time {0:.2f}: {1} is created at {2}".format(self.simulator.simulator_time, entity.name,
                                                                      self.name))
            self.exit_source(entity)

        if "distribution" in self.__dict__:
            try:
                interarrival_time = self.distribution(*self.interarrival_time)
            except TypeError:
                try:
                    interarrival_time = self.distribution(self.interarrival_time)
                except TypeError:
                    raise "Wrong types for transfer in time ({1}) and/or distribution ({0}) for scheduling an event".format(
                        self.interarrival_time, self.distribution)
        else:
            interarrival_time = self.interarrival_time

        relative_delay = interarrival_time

        # Schedule event to create next entity according to the interarrival time
        self.simulator.schedule_event_rel(relative_delay, self, "create_entities")

    def exit_source(self, entity, **kwargs):
        """Schedules the event to exit the source and enter the output node.

        Parameters
        ----------
        entity: object
            the target on which a state change is scheduled.
        kwargs:
            kwargs are the keyword arguments that are used to expand the function.

        """
        self.simulator.schedule_event_now(self, "enter_output_node", entity=entity)

    def enter_output_node(self, entity, **kwargs):
        """Combine the entity with a Vehicle if an vehicle type is given. Combined or not, it schedules an event
        for exiting the output node.

        Parameters
        ----------
        entity: object
            the target on which a state change is scheduled.
        kwargs:
            kwargs are the keyword arguments that are used to expand the function.
            *vehicle_type: Vehicle class
                subclass of Vehicle on which the entity should travel to the next destination.
            *vehicle_speed: int, optional
                speed of the vehicle.

        """
        if "vehicle_type" in self.kwargs:
            if "vehicle_speed" in self.kwargs:
                vehicle = self.kwargs["vehicle_type"](self.simulator, self.kwargs["vehicle_speed"])
            else:
                vehicle = self.kwargs["vehicle_type"](self.simulator)
            vehicle.entities_on_vehicle.append(entity)
            logger.info("Time {0:.2f}: {1} loaded on {2}".format(self.simulator.simulator_time, entity.name,
                                                                 vehicle.name))
            entity = vehicle

        self.simulator.schedule_event_now(self, "exit_output_node", entity=entity)

    def exit_output_node(self, entity, **kwargs):
        """Exit the resource by selecting a link on which the entity should travel to the next destination by
        weighted choice.

        Parameters
        ----------
        entity: object
            the target on which a state change is scheduled.
        kwargs:
            kwargs are the keyword arguments that are used to expand the function.

        """
        try:
            # Selection based on weights in links
            next_list = self.next if isinstance(self.next, list) else [self.next]
            weights = [link.selection_weight for link in next_list]
            link_by_weight = np.random.choice(np.array(next_list), p=weights / np.sum(weights))
            link_by_weight.enter_input_node(entity)
        except AttributeError:
            try:
                if len(next_list) > 1:
                    next_process = np.random.choice(np.array(next_list))
                    next_process.enter_input_node(entity)
                if len(next_list) == 1:
                    self.next.enter_input_node(entity)
            except AttributeError:
                raise AttributeError("{0} has no next process assigned".format(self.name))



"""
Created on: 26-1-2023 13:47

@author: IvS
"""

import pytest

from pydsol.model.entities import Entity, Vehicle


def test_entity(mocker):
    mock_simulator = mocker.Mock()
    mock_time = 1

    entity = Entity(mock_simulator, mock_time)

    assert entity.simulator == mock_simulator
    assert entity.time_creations == mock_time
    assert entity.id == 1
    assert entity.name == "Entity 1"

    entity_2 = Entity(mock_simulator, mock_time, speed=2, test=0, more_test=0)

    assert entity_2.id == 2
    assert entity_2.speed == 2
    assert entity_2.kwargs == dict(test=0, more_test=0)


def test_vehicle(mocker):
    mock_simulator = mocker.Mock()

    vehicle = Vehicle(mock_simulator)

    assert isinstance(vehicle.entities_on_vehicle, list)
    assert vehicle.speed == 10
    assert len(vehicle.entities_on_vehicle) == 0

    vehicle.entities_on_vehicle = ["1", "2", "3"]
    assert vehicle.entities_on_vehicle == ["1", "2", "3"]


if __name__ == '__main__':
    pytest.main()



"""
Created on: 26-1-2023 14:41

@author: IvS
"""

import pytest

from pydsol.model.node import Node


def test_node_class(mocker):
    mock_simulator = mocker.Mock()

    node = Node(mock_simulator)

    assert node.simulator == mock_simulator
    assert node.id == 1
    assert node.name == "Node 1"

    node_2 = Node(mock_simulator, capacity=3, name='test', test=0)

    assert node_2.id == 2
    assert node_2.name == 'test'
    assert node_2.kwargs == dict(name='test', test=0)


def test_enter_input_node(mocker):
    mock_simulator = mocker.Mock()
    mock_entity = mocker.Mock()

    node = Node(mock_simulator)

    node.enter_input_node(mock_entity)


def test_exit_output_node(mocker):
    mock_simulator = mocker.Mock()
    mock_entity = mocker.Mock()

    node = Node(mock_simulator)
    node_2 = Node(mock_simulator)

    # No next
    with pytest.raises(AttributeError):
        node.exit_output_node(mock_entity)

    # Test with only one next node
    node.next = node_2
    node.exit_output_node(mock_entity)

    # Test with two next nodes
    node_3 = Node(mock_simulator)
    node.next = [node_2, node_3]
    node.exit_output_node(mock_entity)

    # Test with two nodes and selection weight
    node_4 = Node(mock_simulator, selection_weight=0.5)
    node_5 = Node(mock_simulator, selection_weight=0.9)

    node.next = [node_4, node_5]
    node.exit_output_node(mock_entity)

    # Test the vehicle
    mock_simulator.simulator_time = 1
    mock_entity.name = "test"

    vehicle = mocker.Mock()
    vehicle.name = "vehicle"

    vehicle.entities_on_vehicle = []
    node_6 = Node(mock_simulator, vehicle_type=vehicle)

    node_6.next = [node_4, node_5]
    node_6.exit_output_node(mock_entity)

    # Test with vehicle speed
    vehicle_speed = 5
    node_7 = Node(mock_simulator, vehicle_type=vehicle, vehicle_speed=vehicle_speed)

    node_7.next = [node_4, node_5]
    node_7.exit_output_node(mock_entity)

    assert isinstance(vehicle_speed, (int, float))


if __name__ == '__main__':
    pytest.main()



""" """
"""
Created on: 22-7-2021 10:22

@author: IvS
"""
import itertools
from pydsol.model.basic_logger import get_module_logger

logger = get_module_logger(__name__)

__all__ = ["Entity", "Vehicle"]


class Entity(object):
    """This class defines a basic entity for a discrete event simulation model."""
    id_iter = itertools.count(1)

    def __init__(self, simulator, t, speed=1.4, **kwargs):
        """

        Parameters
        ----------
        simulator: Simulator object
            simulator of the model.
        speed:
            speed of the entity. Default is 1.4 m/s.
        kwargs:
             kwargs are the keyword arguments that are used to expand the entity class.
             * source_entity: object
                source of which the entity should exit after creation. If keyword is not existing, the entity has no
                source to exit from.
        """
        self.simulator = simulator
        self.time_creations = t
        self.id = next(self.id_iter)
        self.name = "{0} {1}".format(self.__class__.__name__, str(self.id))
        self.speed = speed
        self.kwargs = kwargs


class Vehicle(Entity):
    """This class defines a vehicle in a discrete event simulation model, and is a subclass of Entity. A vehicle is an
    entity of which other entities can be transported. The entities on the vehicle are tracked by a list."""

    def __init__(self, simulator, speed=10, **kwargs):
        """

        Parameters
        ----------
        simulator: Simulator object
            simulator of the model.
        speed:
            speed of the vehicle.
        kwargs:
             kwargs are the keyword arguments that are used to expand the vehicle class.
        """
        super().__init__(simulator, t=simulator.simulator_time, speed=speed, **kwargs)
        self.entities_on_vehicle = []



""" """
"""
Created on: 22-7-2021 10:20

@author: IvS
"""
import itertools
from pydsol.model.basic_logger import get_module_logger
logger = get_module_logger(__name__)

__all__ = ["Sink"]


class Sink(object):
    """This class defines a sink for a discrete event simulation model. A Sink is the end-station of an entity,
    meaning the system will end here for the entity. The entity will be destroyed to reduce the number
    of objects in the model."""
    id_iter = itertools.count(1)

    def __init__(self, simulator, transfer_in_time=0, **kwargs):
        """

        Parameters
        ----------
        simulator: Simulator object
            simulator of the model.
        transfer_in_time: [float, int]
            time it takes to transfer an object into the sink. Default is 0.
        kwargs:
            kwargs are the keyword arguments that are used to expand the sink class.
            *name: str
                user-specified name for the sink.
        """
        self.simulator = simulator
        self.transfer_in_time = transfer_in_time
        if "distribution" in kwargs:
            self.distribution = kwargs["distribution"]

        self.id = next(self.id_iter)
        self.name = "{0} {1}".format(self.__class__.__name__, str(self.id))
        if "name" in kwargs:
            self.name = kwargs["name"]

        self.kwargs = kwargs

        self.processing_entity = None

    def enter_input_node(self, entity):
        """Schedules the event to transfer into the sink and exit the input node.

        Parameters
        ----------
        entity: object
            the target on which a state change is scheduled.

        """
        self.processing_entity = entity

        if "distribution" in self.__dict__:
            try:
                transfer_in_time_dist = self.distribution(*self.transfer_in_time)
            except TypeError:
                try:
                    transfer_in_time_dist = self.distribution(self.transfer_in_time)
                except TypeError:
                    raise "Wrong types for transfer in time ({1}) and/or distribution ({0}) for scheduling an event".format(
                        self.transfer_in_time, self.distribution)
        else:
            transfer_in_time_dist = self.transfer_in_time

        self.simulator.schedule_event_rel(transfer_in_time_dist, self, "exit_input_node", entity=entity)

    def exit_input_node(self, entity, **kwargs):
        """Schedules the event to exit the input node and to destroy the entity.

        Parameters
        ----------
        entity: object
            the target on which a state change is scheduled.

        """
        self.simulator.schedule_event_now(self, "destroy_entity", entity=entity)

    def destroy_entity(self, entity, **kwargs):
        """Destroys the entity.

        Parameters
        ----------
        entity: object
            the target on which a state change is scheduled.

        """
        logger.info("Time {0:.2f}: {1} is destroyed".format(self.simulator.simulator_time, entity.name))
        # To reduce number of objects in model
        del entity



""" """
"""
Created on: 5-8-2021 15:10

@author: IvS
"""
import numpy as np
import itertools

from pydsol.model.queue_model import QueueModel
from pydsol.model.resource import Resource
from pydsol.model.entities import Vehicle

from pydsol.model.basic_logger import get_module_logger

logger = get_module_logger(__name__)

__all__ = ["Server"]


class Server(object):
    """This class defines a basic server for a discrete event simulation. It includes input nodes, input queue,
    multiple resource processes (so seize-and-release principle) based on the capacity, and output node. Since each
    resource process in the server has the same characteristics, the capacity is regulated via a dictionary that
    keeps track of the state of each resource."""
    id_iter = itertools.count(1)

    def __init__(self, simulator, capacity=1, distribution=np.random.triangular, processing_time=(0.1, 0.2, 0.3),
                 transfer_in_time=0, **kwargs):
        """

        Parameters
        ----------
        simulator: Simulator object
            simulator of the model.
        capacity: int
            capacity of the server, i.e., the number of objects that can be handled simultaneously.
        distribution:
            distribution for which the processing time is drawn. Default is np.random.triangular.
        processing_time: int or tuple
            time it takes for the resource to process an object. Default is (0.1, 0.2, 0.3).
        transfer_in_time: int
            time it takes to transfer an object into the resource. Default is 0.
        kwargs:
            kwargs are the keyword arguments that are used to expand the server class.
            *name: str
                user-specified name for the server.

        """
        self.simulator = simulator
        self.next = None

        self.id = next(self.id_iter)
        self.name = "{0} {1}".format(self.__class__.__name__, str(self.id))
        if "name" in kwargs:
            self.name = kwargs["name"]

        self.processing_time = processing_time
        self.distribution = distribution

        self.capacity = capacity
        self.transfer_in_time = transfer_in_time
        self.kwargs = kwargs

        # Queue
        self.input_queue = QueueModel(self)
        self.resources = []
        for i in range(1, self.capacity + 1):
            # TODO make distribution and processing time for multiple resources (in list)
            resource = Resource(i, self.simulator, self.input_queue, self.distribution, self.processing_time,
                                self.transfer_in_time, **kwargs)
            resource.server = self

            self.resources.append(resource)

    def enter_input_node(self, entity, **kwargs):
        """Schedules the event to seize the resource when an object (often an entity) enters the input node.
        If the entity is Vehicle, there are entities on the vehicle and therefore, the vehicle will be separated from
        the entity. The individual entities will try to seize a resource (capacity) in the server.

        Parameters
        ----------
        entity: object
            the target on which a state change is scheduled.
        kwargs:
            kwargs are the keyword arguments that are used to expand the function.

        """
        if isinstance(entity, Vehicle):
            vehicle = entity
            for entity in vehicle.entities_on_vehicle:
                logger.info("Time {0:.2f}: {1} loaded off {2}".format(self.simulator.simulator_time, entity.name,
                                                                      vehicle.name))
                self.simulator.schedule_event_now(self, "seize_resource", entity=entity)
                # To reduce number of objects in model
            del vehicle
        else:
            self.simulator.schedule_event_now(self, "seize_resource", entity=entity)


    def seize_resource(self, entity, **kwargs):
        """Process to seize a resource by the entity. If the resource is free and there is no queue, the entity
        can directly seize the resource. If the resource is free and there is a queue, the first entity of the queue
        seizes the resource and this entity is added to the queue. If the resource is occupied, the entity is directly
        added to the queue.

        Parameters
        ----------
        entity: Entity
            the target on which a state change is scheduled.
        kwargs:
            kwargs are the keyword arguments that are used to expand the function.

        """
        not_seized = [resource for resource in self.resources if not resource.resource_seized]

        if len(not_seized) == 0:
            self.input_queue.contents.append(entity)
            logger.debug(
                "Time {0:.2f}: {1} added to the queue".format(self.simulator.simulator_time, entity.name))
        else:
            resource_to_seize = not_seized[0]
            if len(self.input_queue.contents) == 0:
                resource_to_seize.resource_seized = True
                resource_to_seize.processing_entity = entity
                self.simulator.schedule_event_now(resource_to_seize, "exit_input_node")
                logger.debug(
                    "Time {0:.2f}: {1} seized {2}".format(self.simulator.simulator_time, entity.name,
                                                          resource_to_seize.name))
            else:
                resource_to_seize.resource_seized = True
                resource_to_seize.processing_entity = self.input_queue.contents[0]
                self.simulator.schedule_event_now(resource_to_seize, "exit_input_node")
                logger.debug(
                    "Time {0:.2f}: {1} seized {2}".format(self.simulator.simulator_time,
                                                          resource_to_seize.processing_entity.name, resource_to_seize.name))
                self.input_queue.contents.pop(0)
                self.input_queue.contents.append(entity)

    def enter_output_node(self, entity, **kwargs):
        """Combine the entity with a Vehicle if an vehicle type is given. Combined or not, it schedules an event
        for exiting the output node.

        Parameters
        ----------
        entity: Entity
            the target on which a state change is scheduled.
        kwargs:
            kwargs are the keyword arguments that are used to expand the function.
            *vehicle_type: Vehicle class
                subclass of Vehicle on which the entity should travel to the next destination.
            *vehicle_speed: int, optional
                speed of the vehicle.

        """
        # If there is a vehicle
        if "vehicle_type" in self.kwargs:
            if "vehicle_speed" in self.kwargs:
                vehicle = self.kwargs["vehicle_type"](self.simulator, vehicle_speed=self.kwargs["vehicle_speed"])
            else:
                vehicle = self.kwargs["vehicle_type"](self.simulator)
            vehicle.entities_on_vehicle.append(entity)
            logger.info("Time {0:.2f}: {1} loaded on {2}".format(self.simulator.simulator_time, entity.name,
                                                                 vehicle.name))
            entity = vehicle

        self.simulator.schedule_event_now(self, "exit_output_node", entity=entity)

    def exit_output_node(self, entity, **kwargs):
        """Exit the resource by selecting a link on which the entity should travel to the next destination by
        weighted choice.

        Parameters
        ----------
        entity: Entity
            the target on which a state change is scheduled.
        kwargs:
            kwargs are the keyword arguments that are used to expand the function.

        """
        try:
            # Selection based on weights in links
            next_list = self.next if isinstance(self.next, list) else [self.next]
            weights = [link.selection_weight for link in next_list]
            link_by_weight = np.random.choice(np.array(next_list), p=weights / np.sum(weights))
            link_by_weight.enter_input_node(entity)
        except AttributeError:
            try:
                if len(next_list) > 1:
                    next_process = np.random.choice(np.array(next_list))
                    next_process.enter_input_node(entity)
                if len(next_list) == 1:
                    self.next.enter_input_node(entity)
            except AttributeError:
                raise AttributeError("{0} has no next process assigned".format(self.name))



""" """
"""
Created on: 23-7-2021 16:00

@author: IvS
"""
import math
import numpy as np
import itertools

from pydsol.model.basic_logger import get_module_logger
from pydsol.model.entities import Entity

logger = get_module_logger(__name__)

__all__ = ["Node"]


class Node(object):
    """This class defines a basic node for travelling in a discrete event simulation model. Input and output links
    are assigned by the link."""
    id_iter = itertools.count(1)

    def __init__(self, simulator, capacity=math.inf, **kwargs):
        """

        Parameters
        ----------
        simulator: Simulator object
            simulator of the model.
        capacity: int
            number of travellers that are simultaneously allowed on the node. Default is infinity.
        kwargs:
            kwargs are the keyword arguments that are used to expand the node class.
            *name: str
                user-specified name for the node.
        """
        self.simulator = simulator
        self.capacity = capacity

        self.next = None

        self.id = next(self.id_iter)
        self.name = "{0} {1}".format(self.__class__.__name__, str(self.id))
        if "name" in kwargs:
            self.name = kwargs["name"]
        self.kwargs = kwargs

    def enter_input_node(self, entity: Entity, **kwargs):
        self.simulator.schedule_event_now(self, "exit_output_node", entity=entity)

    def exit_output_node(self, entity: Entity, **kwargs):
        """Exit the resource by selecting a link on which the entity should travel to the next destination by
        weighted choice.

        Parameters
        ----------
        entity: Entity
            the target on which a state change is scheduled.
        kwargs:
            kwargs are the keyword arguments that are used to expand the function.

        """
        # If there is a vehicle
        if "vehicle_type" in self.kwargs:
            if "vehicle_speed" in self.kwargs:
                vehicle = self.kwargs["vehicle_type"](self.simulator, vehicle_speed=self.kwargs["vehicle_speed"])
            else:
                vehicle = self.kwargs["vehicle_type"](self.simulator)
            vehicle.entities_on_vehicle.append(entity)
            logger.info("Time {0:.2f}: {1} loaded on {2} at {3}".format(self.simulator.simulator_time, entity.name,
                                                                        vehicle.name, self.name))
            entity = vehicle

        try:
            # Selection based on weights in links
            next_list = self.next if isinstance(self.next, list) else [self.next]
            weights = [link.selection_weight for link in next_list]
            link_by_weight = np.random.choice(np.array(next_list), p=weights / np.sum(weights))
            link_by_weight.enter_input_node(entity)
        except AttributeError:
            try:
                if len(next_list) > 1:
                    next_process = np.random.choice(np.array(next_list))
                    next_process.enter_input_node(entity)
                if len(next_list) == 1:
                    self.next.enter_input_node(entity)
            except AttributeError:
                raise AttributeError("{0} has no next process assigned".format(self.name))



"""
Created on: 30-1-2023 14:22

@author: IvS
"""
import pytest
import numpy as np

from pydsol.model.resource import Resource


def test_resource_class(mocker):
    mock_simulator = mocker.Mock()
    mock_queue = mocker.Mock(contents=[], attached_to=mocker.Mock())

    resource = Resource(1, mock_simulator, mock_queue, np.random.triangular, (0.1, 0.2, 0.3), 0)
    assert resource.name == "Resource 1"

    resource = Resource(5, mock_simulator, mock_queue, np.random.triangular, (0.1, 0.2, 0.3), 0)
    assert resource.name == "Resource 5"


def test_exit_input_node(mocker):
    mock_simulator = mocker.Mock(simulator_time=1)
    mock_queue = mocker.Mock(contents=[], attached_to=mocker.Mock())
    mock_entity = mocker.Mock(name="entity1")

    resource = Resource(1, mock_simulator, mock_queue, np.random.triangular, (0.1, 0.2, 0.3), 0)
    resource.processing_entity = mock_entity
    assert isinstance(resource.processing_entity, object)
    resource.exit_input_node()


def test_enter_resource(mocker):
    mock_simulator = mocker.Mock(simulator_time=1)
    mock_queue = mocker.Mock(contents=[], attached_to=mocker.Mock())
    mock_entity = mocker.Mock(name="entity1")

    resource = Resource(1, mock_simulator, mock_queue, np.random.triangular, (0.1, 0.2, 0.3), 0)
    resource.processing_entity = mock_entity
    assert isinstance(resource.processing_entity, object)
    assert isinstance(resource.transfer_in_time, (int, float))
    resource.enter_resource()


def test_processing(mocker):
    mock_simulator = mocker.Mock(simulator_time=1)
    mock_queue = mocker.Mock(contents=[], attached_to=mocker.Mock())
    mock_entity = mocker.Mock(name="entity1")

    # Basic
    resource = Resource(1, mock_simulator, mock_queue, np.random.triangular, (0.1, 0.2, 0.3), 0)
    resource.processing_entity = mock_entity
    resource.processing()

    # With wrong distribution and unit
    with pytest.raises(TypeError):
        resource = Resource(1, mock_simulator, mock_queue, np.random.random, (0.1, 0.2, 0.3), 0)
        resource.processing_entity = mock_entity
        resource.processing()

    # With distribution and wrong unit
    with pytest.raises(TypeError):
        resource = Resource(1, mock_simulator, mock_queue, np.random.triangular, 1, 0)
        resource.processing_entity = mock_entity
        resource.processing()

    # With no distribution
    resource = Resource(1, mock_simulator, mock_queue, distribution=None, processing_time=1, transfer_in_time=0)
    resource.processing_entity = mock_entity
    resource.processing()

    # With no distribution but wrong unit processing time
    with pytest.raises(TypeError):
        resource = Resource(1, mock_simulator, mock_queue, distribution=None, processing_time=(0, 1, 3),
                            transfer_in_time=0)
        resource.processing_entity = mock_entity
        resource.processing()

    # With no processing time
    with pytest.raises(TypeError):
        resource = Resource(1, mock_simulator, mock_queue, np.random.triangular, None, 0)
        resource.processing_entity = mock_entity
        resource.processing()

    # Test processing entity in kwargs
    resource = Resource(1, mock_simulator, mock_queue, np.random.triangular, (0.1, 0.2, 0.3), 0)
    resource.processing(processing_entity=mock_entity)
    assert resource.processing_entity == mock_entity


def test_exit_resource(mocker):
    mock_simulator = mocker.Mock(simulator_time=1)
    mock_queue = mocker.Mock(contents=[], attached_to=mocker.Mock())
    mock_entity = mocker.Mock(name="entity1")

    resource = Resource(1, mock_simulator, mock_queue, np.random.triangular, (0.1, 0.2, 0.3), 0)
    resource.processing_entity = mock_entity
    resource.server = mocker.Mock(name="server1")

    assert isinstance(resource.processing_entity, object)
    assert isinstance(resource.server, object)

    resource.exit_resource()


def test_check_queue(mocker):
    mock_simulator = mocker.Mock(simulator_time=1)
    mock_queue = mocker.Mock(contents=[], attached_to=mocker.Mock())
    mock_entity = mocker.Mock(name="entity1")

    resource = Resource(1, mock_simulator, mock_queue, np.random.triangular, (0.1, 0.2, 0.3), 0)

    # When queue is zero
    resource.check_queue()
    assert len(resource.input_queue.contents) == 0
    assert not resource.resource_seized

    # When there is a queue get the first item
    resource.input_queue.contents = [mock_entity]
    assert len(resource.input_queue.contents) == 1

    resource.check_queue()
    assert resource.resource_seized
    assert resource.processing_entity == mock_entity
    assert len(resource.input_queue.contents) == 0


if __name__ == '__main__':
    pytest.main()



""" """
"""
Created on: 9-8-2021 17:02

@author: IvS
"""
import itertools

from pydsol.model.basic_logger import get_module_logger
from pydsol.model.entities import Entity

logger = get_module_logger(__name__)

__all__ = ["Resource"]


class Resource(object):
    """This class defines a basic resource for a discrete event simulation. An object (often an entity) can seize
    a resource. If the resource is seized, other objects cannot seize it anymore until it is released after the
    processing time. The remaining objects are put into a queue."""
    id_iter = itertools.count(1)

    def __init__(self, id, simulator, queue, distribution, processing_time, transfer_in_time, **kwargs):
        """

        Parameters
        ----------
        simulator: Simulator object
            simulator of the model.
        queue: Queue object
            queue where the remaining objects are places when the resource is seized.
        distribution:
            distribution from which the processing time is drawn
        processing_time: Union[int, str]
            time it takes for the resource to process an object. Default is np.random.triangular(0.1, 0.2, 0.3).
        transfer_in_time: int
            time it takes to transfer an object into the resource. Default is 0.
        kwargs:
            kwargs are the keyword arguments that are used to expand the resource class.
            *name: str
                user-specified name for the resource.

        """
        self.simulator = simulator

        self.id = id
        self.name = "{0} {1}".format(self.__class__.__name__, str(self.id))

        self.distribution = distribution
        self.processing_time = processing_time
        self.transfer_in_time = transfer_in_time
        self.kwargs = kwargs

        self.server = object

        # Create queue
        self.input_queue = queue

        self.resource_seized = False

        self.processing_entity = None

    def exit_input_node(self):
        """Schedules the event to enter the resource when the input node is exited. To ensure that the right
                resource is called, the resource key is given.

                Parameters
                ----------
                entity: Entity
                    the target on which a state change is scheduled.
                """
        logger.debug(
            "Time {0:.2f}: {1} seized {2}".format(self.simulator.simulator_time, self.processing_entity.name, self.name))

        self.simulator.schedule_event_now(self, "enter_resource")

    def enter_resource(self):
        """Schedules the event to transfer into the resource and starts processing.

        Parameters
        ----------
        entity: object
            the target on which a state change is scheduled.
        """
        self.simulator.schedule_event_rel(self.transfer_in_time, self, "processing")

    def processing(self, **kwargs):
        """Schedules the event to process the entity and exit the resource.

        Parameters
        ----------
        entity: object
            the target on which a state change is scheduled.
        kwargs:
            kwargs are the keyword arguments that are used to expand the function.

        """
        processing_time_dist = self.set_processing_time()

        if "processing_entity" in kwargs:
            self.processing_entity = kwargs["processing_entity"]

        self.simulator.schedule_event_rel(processing_time_dist, self, "exit_resource", **kwargs)

    def set_processing_time(self):
        """Draws a processing time from the distribution and parameter processing time.
        """
        if (self.distribution is None) & (isinstance(self.processing_time, (int, float))):
            processing_time_dist = self.processing_time
        elif isinstance(self.processing_time, (int, float)):
            processing_time_dist = self.distribution(self.processing_time)
        else:
            try:
                processing_time_dist = self.distribution(*self.processing_time)
            except TypeError:
                raise "Wrong types for processing time ({1}) and/or distribution ({0}) for scheduling an event".format(
                    self.processing_time, self.distribution)
        processing_time_dist = max(0, processing_time_dist)
        return processing_time_dist

    def exit_resource(self, **kwargs):
        """Exits the resource. The resource is released and the queue attached to this resource is checked.

        Parameters
        ----------
        entity:
            the target on which a state change is scheduled.
        kwargs:
            kwargs are the keyword arguments that are used to expand the function.

        """
        logger.info(
            "Time {0:.2f}: {1} is processed by {2} of {3}".format(self.simulator.simulator_time, self.processing_entity,
                                                                  self.name, self.server.name))
        self.simulator.schedule_event_now(self.server, "enter_output_node", entity=self.processing_entity)

        # Release resource and check queue
        self.simulator.schedule_event_now(self, "check_queue", priority=1)

    def check_queue(self, **kwargs):
        """Checks whether there are objects in the queue to seize the resource. If there are no objects in the queue,
        the resource stays free. If there are objects in the queue, the first object seizes the resource.

        Parameters
        ----------
        obj:
            optional
        kwargs:
            kwargs are the keyword arguments that are used to expand the function.

        """
        logger.info(
            "Time {0:.2f}: Length of queue attached to {1} is {2}".format(self.simulator.simulator_time, self.name,
                                                                          len(self.input_queue.contents)))
        if len(self.input_queue.contents) == 0:
            self.resource_seized = False
            pass
        else:
            self.resource_seized = True
            self.processing_entity = self.input_queue.contents[0]
            self.simulator.schedule_event_now(self, "exit_input_node")
            self.input_queue.contents.pop(0)





"""
Created on: 30-1-2023 14:59

@author: IvS
"""
import pytest
import numpy as np

from pydsol.model.sink import Sink


def test_sink_class(mocker):
    mock_simulator = mocker.Mock()

    sink = Sink(mock_simulator)

    assert isinstance(sink.transfer_in_time, (int, float))
    assert sink.id == 1
    assert sink.name == "Sink 1"
    assert isinstance(sink.name, str)

    sink_2 = Sink(mock_simulator, transfer_in_time=(0, 1, 2), distribution=np.random.triangular, name="test")

    assert sink_2.id == 2
    assert isinstance(sink_2.name, str)


def test_enter_input_node(mocker):
    mock_simulator = mocker.Mock()
    mock_entity = mocker.Mock()

    # Iterable distribution
    sink = Sink(mock_simulator, transfer_in_time=(0, 1, 2), distribution=np.random.triangular)
    sink.enter_input_node(mock_entity)

    # None-iterable distribution
    sink = Sink(mock_simulator, transfer_in_time=1, distribution=np.random.normal)
    sink.enter_input_node(mock_entity)

    # Not matching distribution and transfer_in_time
    with pytest.raises(TypeError):
        sink = Sink(mock_simulator, transfer_in_time=1, distribution=np.random.triangular)
        sink.enter_input_node(mock_entity)

    # Only tranfer time
    sink = Sink(mock_simulator, transfer_in_time=1)
    sink.enter_input_node(mock_entity)


def test_exit_output_node(mocker):
    mock_simulator = mocker.Mock()
    mock_entity = mocker.Mock()

    sink = Sink(mock_simulator)
    sink.exit_input_node(mock_entity)
    assert isinstance(mock_entity, object)


def test_destroy_entity(mocker):
    mock_simulator = mocker.Mock(simulator_time=1)
    mock_entity = mocker.Mock(name="entity")

    sink = Sink(mock_simulator)
    sink.destroy_entity(mock_entity)


if __name__ == '__main__':
    pytest.main()



""" """
"""
Created on: 22-7-2021 13:42

@author: IvS
"""
import itertools
from pydsol.model.basic_logger import get_module_logger

logger = get_module_logger(__name__)

__all__ = ["Link", "TimePath"]


class Link(object):
    """This class defines a basic link between an origin and destination for a model."""
    id_iter = itertools.count(1)

    def __init__(self, simulator, origin, destination, length, selection_weight=1, **kwargs):
        """

        Parameters
        ----------
        simulator: Simulator object
            simulator of the model.
        origin: object
            origin of the link, i.e., 'from'.
        destination: object
            destination of the link, i.e., 'to'.
        length: int
            length of the link in a distance unit.
        selection_weight: float
            weight expression for the link when selecting a link.
        kwargs:
            kwargs are the keyword arguments that are used to expand the link class.
            *name: str
                user-specified name for the link
        """
        self.simulator = simulator
        self.id = next(self.id_iter)
        self.name = "{0} {1}".format(self.__class__.__name__, str(self.id))
        if "name" in kwargs:
            self.name = kwargs["name"]

        self.next = None  # is next item

        self.origin = origin
        self.destination = destination

        self.length = length
        self.selection_weight = selection_weight

        self.kwargs = kwargs

    def enter_input_node(self, entity, **kwargs):
        self.enter_link(entity)

    def enter_link(self, entity, **kwargs):
        """Determines the time for travelling on the link by length and speed of the entity, and
        schedules the event for exiting the link.

        Parameters
        ----------
        entity: object
            the target on which a state change is scheduled.
        kwargs:
            kwargs are the keyword arguments that are used to expand the link class.

        """
        relative_delay = self.length / entity.speed

        self.simulator.schedule_event_rel(relative_delay, self, "exit_link", entity=entity)

        logger.info("Time {0:.2f}: {1} enters {2} from {3} to {4}".format(self.simulator.simulator_time,
                                                                          entity.name, self.name,
                                                                          self.origin.name, self.destination.name))

    def exit_link(self, entity, **kwargs):
        """Schedules the event for entering the input node of the destination of the link. Each destination needs
        a method "enter_input_node".

        Parameters
        ----------
        entity: object
            the target on which a state change is scheduled.
        kwargs:
            kwargs are the keyword arguments that are used to expand the link class.

        """
        self.next.enter_input_node(entity)


class TimePath(object):
    """This class defines a time path between an origin and destination for a model."""
    id_iter = itertools.count(1)

    def __init__(self, simulator, origin, destination, time, selection_weight=1, **kwargs):
        """

        Parameters
        ----------
        simulator: Simulator object
            simulator of the model.
        origin: object
            origin of the path, i.e., 'from'.
        destination: object
            destination of the path, i.e., 'to'.
        time: [int, float]
            distance of path in time unit.
        selection_weight: float
            weight expression for the path when selecting a path.
        kwargs:
            kwargs are the keyword arguments that are used to expand the path class.
            *name: str
                user-specified name for the path
        """
        self.simulator = simulator
        self.id = next(self.id_iter)
        self.name = "{0} {1}".format(self.__class__.__name__, str(self.id))
        if "name" in kwargs:
            self.name = kwargs["name"]

        self.next = None  # is next item

        self.origin = origin
        self.destination = destination

        self.time = time
        self.selection_weight = selection_weight

        self.kwargs = kwargs

    def enter_input_node(self, entity, **kwargs):
        self.enter_path(entity)

    def enter_path(self, entity, **kwargs):
        """Determines the time for travelling on the link by length and speed of the entity, and
        schedules the event for exiting the link.

        Parameters
        ----------
        entity: object
            the target on which a state change is scheduled.
        kwargs:
            kwargs are the keyword arguments that are used to expand the link class.

        """
        relative_delay = self.time

        self.simulator.schedule_event_rel(relative_delay, self, "exit_path", entity=entity)

        logger.info("Time {0:.2f}: {1} enters {2} from {3} to {4}".format(self.simulator.simulator_time,
                                                                          entity.name, self.name,
                                                                          self.origin.name, self.destination.name))

    def exit_path(self, entity, **kwargs):
        """Schedules the event for entering the input node of the destination of the link. Each destination needs
        a method "enter_input_node".

        Parameters
        ----------
        entity: object
            the target on which a state change is scheduled.
        kwargs:
            kwargs are the keyword arguments that are used to expand the link class.

        """
        self.next.enter_input_node(entity)



from setuptools import setup, find_packages

import pathlib

# The directory containing this file
HERE = pathlib.Path(__file__).parent
REQUIREMENTS = (HERE / "requirements.txt").read_text().splitlines()

# The text of the README file
README = (HERE / "README.rst").read_text()

setup(
    name="pydsol-model",
    version="0.1",
    long_description=README,
    long_description_content_type='text/x-rst',
    author="Isabelle M. van Schilt",
    url="https://github.com/imvs95/pydsol-model",
    packages=["pydsol", "pydsol.model"],
    package_dir={'pydsol': 'src/pydsol', "pydsol.model": "src/pydsol/model"},
    install_requires=REQUIREMENTS
)



""" """
"""
Created on: 6-8-2021 15:38

@author: IvS
"""

from pydsol.model.basic_logger import get_module_logger

logger = get_module_logger(__name__)

__all__ = ["QueueModel"]

class QueueModel(object):
    """This class defines a basic queue  to a specific object for a discrete event simulation model,
    and keeps track of the objects in the queue with a contents list."""
    def __init__(self, attached_to):
        """

        Parameters
        ----------
        attached_to: object
            object where the queue is attached to, for example a Server.
        """
        self.attached_to = attached_to
        self.contents = []



""" """
"""
Created on: 25-6-2021 14:43

@author: IvS
"""
import logging
import sys

__all__ = ["get_module_logger"]

def get_module_logger(mod_name, level=logging.CRITICAL):
    logger = logging.getLogger(mod_name)
    handler = logging.StreamHandler(stream=sys.stdout)
    msg_format = '%(asctime)s [%(levelname)s] %(message)s (%(name)s - %(filename)s: line %(lineno)s)'
    date_format = '%Y-%m-%d %H:%M:%S'
    formatter = logging.Formatter(fmt=msg_format, datefmt=date_format)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(level)
    logger.propagate = False
    return logger
